# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Ryan Robison
# Date: 2013mar20

import gpi
from gpi import QtWidgets

# WIDGET
class SHAPES_GROUP(gpi.GenericWidgetGroup):
    """A combination of Sliders and SpinBoxes to form a
    unique widget suitable for defining the filter shape.
    """
    valueChanged = gpi.Signal()

    def __init__(self, title, parent=None):
        super().__init__(title, parent)

        self._val = {}
        self._val['size'] = 1
        self._val['width'] = 1

        self.sb = gpi.BasicSpinBox()  # size
        self.sb.set_label('size:')
        self.sb.set_min(1)
        self.sb.set_val(1)
        self.sb.set_max(gpi.GPI_INT_MAX)

        self.slb = gpi.BasicSlider()  # width
        self.slb.set_min(1)
        self.slb.set_val(1)
        self.slb.set_max(1)

        self.slider_label = QtWidgets.QLabel('width:')

        self.sb.valueChanged.connect(self.sizeChange)
        self.slb.valueChanged.connect(self.widthChange)

        hbox = QtWidgets.QHBoxLayout()
        hbox.addWidget(self.slider_label)
        hbox.addWidget(self.slb)
        hbox.setStretch(0, 0)
        hbox.setSpacing(5)

        hboxGroup = QtWidgets.QHBoxLayout()
        hboxGroup.addWidget(self.sb)
        hboxGroup.addLayout(hbox)
        hboxGroup.setStretch(0, 0)
        hboxGroup.setContentsMargins(0, 0, 0, 0)  # we don't need margins here
        hboxGroup.setSpacing(30)
        self.setLayout(hboxGroup)

    # setters
    def set_val(self, val):
        """A python-dict containing: size, and width parms. """
        sig = False
        if 'size' in val:
            self._val['size'] = val['size']
            self.setSizeQuietly(val['size'])
            sig = True
        if 'width' in val:
            self._val['width'] = val['width']
            self.setWidthQuietly(val['width'])
            sig = True
        if sig:
            self.valueChanged.emit()

    # getters
    def get_val(self):
        return self._val

    # support
    def sizeChange(self, val):
        self._val['size'] = val
        self.setSizeQuietly(val)
        self.valueChanged.emit()

    def widthChange(self, val):
        self._val['width'] = val
        self.setWidthQuietly(val)
        self.valueChanged.emit()

    def setSizeQuietly(self, val):
        self.sb.blockSignals(True)
        self.sb.set_val(val)
        self.slb.set_max(val)
        self.sb.blockSignals(False)

    def setWidthQuietly(self, val):
        self.slb.blockSignals(True)
        self.slb.set_val(val)
        self.slb.blockSignals(False)

# End of WIDGET declaration


class ExternalNode(gpi.NodeAPI):
    """Geometric function generator (used to make different shapes in 1, 2, or 3 dimensions).

    INPUT:  Optional input - data are not used, but the shape of this array (ndim and array shape) are used for the output array

    OUTPUT: 1D, 2D, or 3D real-valued numpy array

    WIDGETS: (some of these parameters change in special-use cases)
    Dimensions - specify whether output data is 1D, 2D, or 3D
    Function - type of Geometric function to create
    Compute - turn off to change parameters without generating new data (turn on to generate data)
    Dimension: i - for the ith dimesion:
                   size is the length of that dimension for the output data
                   width is (usually) the width of the Geometric Function, <= size

                   Note for some Functions, only Dimension: 1 is shown even for 2D and 3D dimensions,
                      implying that the output data will have the same size/width in all dimensions

    Pass Value: Maximum value in function
    Stop Value: Value in data outside of function
    Window: allows one to taper function, creates a linear ramp between the pass and stop value.
            Window units are in %, from 0 (creating a sharp edge)
                                   to 100 (tapering from the edge to the middle of the object)

    Some Special Use Cases:
      For Sinc function, width specifies the width of the main lobe.
                         Window allows for additional tapering of the edges
      For Poly, one can specify the 0th, 1st, and 2nd order coefficients of an N-dimensional polynomial
      For Noise, one can specify the Noise standard deviation.  Everytime the module is poked, it generates a
                 new pseudo-random instance of noise
    """

    def initUI(self):
        # Widgets
        self.dim_base_name = 'Dimension['
        self.ndim = 3  # max of 3 dims for now
        self.addWidget('ExclusivePushButtons', 'Dimensions',
                       buttons=['1D', '2D', '3D'],
                       val=1)
        self.addWidget('ExclusiveRadioButtons', 'Function',
                       buttons=['Ellipse', 'Rectangle', 'Hanning',
                                'Gaussian', 'Sinc', 'Poly', 'Riesz', 'Noise'],
                       val=0)
        self.addWidget(
            'PushButton', 'Compute', button_title='OFF', toggle=True)
        self.addWidget('PushButton', 'Equal Dimensions', toggle=True, val=0)
        for i in range(self.ndim):
            self.addWidget('SHAPES_GROUP', self.dim_base_name+str(-i-1)+']')
        self.addWidget('DoubleSpinBox', 'Pass Value:', val=1.0, decimals=5)
        self.addWidget('DoubleSpinBox', 'Stop Value:', val=0.0, decimals=5)
        self.addWidget('Slider', 'Window:', val=1.0)
        self.addWidget('DoubleSpinBox', 'Poly Const:', val=0.0, decimals=5)
        self.addWidget('DoubleSpinBox', 'Poly i:', val=0.0, decimals=5)
        self.addWidget('DoubleSpinBox', 'Poly i^2:', val=0.0, decimals=5)
        self.addWidget('DoubleSpinBox', 'Poly j:', val=0.0, decimals=5)
        self.addWidget('DoubleSpinBox', 'Poly j^2:', val=0.0, decimals=5)
        self.addWidget('DoubleSpinBox', 'Poly k:', val=0.0, decimals=5)
        self.addWidget('DoubleSpinBox', 'Poly k^2:', val=0.0, decimals=5)
        self.addWidget('DoubleSpinBox', 'Noise Std:', val=1.0, decimals=5)
        self.addWidget('PushButton', 'Real/Complex', toggle=True, val=0)
        self.addWidget('PushButton', 'Fixed Seed', toggle=True, val=0)
        self.addWidget('SpinBox', 'Seed', val = 0)
        
        # IO Ports
        self.addInPort('in', 'NPYarray', obligation = gpi.OPTIONAL)
        self.addOutPort('out', 'NPYarray')

    def validate(self):
        '''update the widgets based on the selected options
        '''

        # If input data is present, set size accordingly
        indat = self.getData('in')
        if indat is not None:
          #if indat.ndim not in [1,2,3]:
          #  self.log.warn('input must be 1D, 2D, or 3D')
          #  return(1)
          #self.setAttr('Dimensions',val=indat.ndim-1)
          #for i in range(indat.ndim):
          for i in range(self.getVal('Dimensions')+1):
              try:
                dval = self.getVal(self.dim_base_name+str(-i-1)+']')
                dval['size'] = indat.shape[-i-1]
                self.setAttr(self.dim_base_name+str(-i-1)+']',quietval=dval)
              except:
                pass

        # GETTING WIDGET INFO
        ndim = self.getVal('Dimensions') + 1
        function = self.getVal('Function')
        same_dim = self.getVal('Equal Dimensions')

        # visibility and bounds
        if self.widgetEvents() or self.portEvents():
            for i in range(self.ndim):
                if i < ndim and (i == 0 or not same_dim):
                    self.setAttr(self.dim_base_name+str(-i-1)+']',
                                      visible=True)
                else:
                    self.setAttr(self.dim_base_name+str(-i-1)+']',
                                      visible=False)

            if function in [5, 7]:
                self.setAttr('Pass Value:', visible=False)
                self.setAttr('Stop Value:', visible=False)
            else:
                self.setAttr('Pass Value:', visible=True)
                self.setAttr('Stop Value:', visible=True)

            if function in [3, 5, 7]:
                self.setAttr('Window:', visible=False)
            else:
                self.setAttr('Window:', visible=True)

            if function == 5:
                self.setAttr('Poly Const:', visible=True)
                self.setAttr('Poly i:', visible=True)
                self.setAttr('Poly i^2:', visible=True)
                if ndim > 1:
                    self.setAttr('Poly j:', visible=True)
                    self.setAttr('Poly j^2:', visible=True)
                else:
                    self.setAttr('Poly j:', visible=False)
                    self.setAttr('Poly j^2:', visible=False)
                if ndim > 2:
                    self.setAttr('Poly k:', visible=True)
                    self.setAttr('Poly k^2:', visible=True)
                else:
                    self.setAttr('Poly k:', visible=False)
                    self.setAttr('Poly k^2:', visible=False)
            else:
                self.setAttr('Poly Const:', visible=False)
                self.setAttr('Poly i:', visible=False)
                self.setAttr('Poly i^2:', visible=False)
                self.setAttr('Poly j:', visible=False)
                self.setAttr('Poly j^2:', visible=False)
                self.setAttr('Poly k:', visible=False)
                self.setAttr('Poly k^2:', visible=False)

            if function == 7:
                self.setAttr('Noise Std:', visible=True)
                self.setAttr('Real/Complex', visible=True)
                self.setAttr('Fixed Seed', visible=True)
                if(self.getVal('Fixed Seed')):
                    self.setAttr('Seed', visible=True)
            else:
                self.setAttr('Noise Std:', visible=False)
                self.setAttr('Real/Complex', visible=False)
                self.setAttr('Fixed Seed', visible=False)
                self.setAttr('Seed', visible=False)

        return(0)

    def compute(self):

        import numpy as np

        # GETTING WIDGET INFO
        compute = self.getVal('Compute')
        function = self.getVal('Function')
        ndim = self.getVal('Dimensions') + 1
        same_dim = self.getVal('Equal Dimensions')
        complex = self.getVal('Real/Complex')
        fixedSeed = self.getVal('Fixed Seed')

        self.dims = {}
        self.widths = {}
        for i in range(self.ndim):
            if not same_dim:
                val = self.getVal(self.dim_base_name+str(-i-1)+']')
            else:
                val = self.getVal(self.dim_base_name+str(-1)+']')

            self.dims[i+1] = val['size']
            self.widths[i+1] = val['width']

        if function in [5, 7]:
            passVal = 1.0
            stopVal = 0.0
        else:
            passVal = self.getVal('Pass Value:')
            stopVal = self.getVal('Stop Value:')

        if function not in [3, 5, 7]:
            window = float(self.getVal('Window:')) / 100

        if function == 5:
            polyC = self.getVal('Poly Const:')
            polyi = self.getVal('Poly i:')
            polyi2 = self.getVal('Poly i^2:')
            polyk = 0.0
            polyk2 = 0.0
            if ndim > 1:
                polyj = self.getVal('Poly j:')
                polyj2 = self.getVal('Poly j^2:')
            else:
                polyj = 0.0
                polyj2 = 0.0
            if ndim > 2:
                polyk = self.getVal('Poly k:')
                polyk2 = self.getVal('Poly k^2:')
            else:
                polyk = 0.0
                polyk2 = 0.0

        if function == 7:
            std = self.getVal('Noise Std:')

        if function == 7 and same_dim:
            for i in range(self.ndim):
                self.dims[i + 1] = self.dims[1]

        # SETTING WIDGET INFO
        if compute:
            self.setAttr('Compute', button_title="ON")
        else:
            self.setAttr('Compute', button_title="OFF")
            
        if complex:
            self.setAttr('Real/Complex', button_title="Complex")
        else:
            self.setAttr('Real/Complex', button_title="Real")
            
        if fixedSeed:
            self.setAttr('Fixed Seed', button_title="ON")
        else:
            self.setAttr('Fixed Seed', button_title="OFF")

        # SETTING PORT INFO
        if compute:
            dim1 = self.dims[1]
            dim2 = self.dims[2]
            dim3 = self.dims[3]
            # ranges are 0 in the middle, and 1. at the index = prescribed width
            range1 = (2 * np.linspace(-(dim1 / 2), (dim1 - 1 - dim1/2), dim1) /
                      self.widths[1])
            range2 = (2 * np.linspace(-(dim2 / 2), (dim2 - 1 - dim2/2), dim2) /
                      self.widths[2])
            range3 = (2 * np.linspace(-(dim3 / 2), (dim3 - 1 - dim3/2), dim3) /
                      self.widths[3])

            # JGP for sinc function windowing
            radnorm = float(self.widths[1])/float(dim1)

            if function != 7:
                if ndim == 1:
                    cart = np.zeros([self.dims[1], 1])
                    cart[:, 0] = range1
                    radius = abs(cart[:, 0])
                    out = stopVal * np.ones(self.dims[1])
                elif ndim == 2:
                    cart = np.zeros([self.dims[2], self.dims[1], 2])
                    cart[:, :, 0] = np.tile(range1, self.dims[2])\
                                      .reshape(self.dims[2], self.dims[1])
                    cart[:, :, 1] = np.tile(range2, self.dims[1])\
                                      .reshape(self.dims[1], self.dims[2])\
                                      .transpose(1, 0)
                    radius = np.sqrt(np.square(cart[:, :, 0]) +
                                     np.square(cart[:, :, 1]))
                    out = stopVal * np.ones([self.dims[2], self.dims[1]])
                else:
                    cart = np.zeros([self.dims[3], self.dims[2],
                                     self.dims[1], 3])
                    cart[:, :, :, 0] = np.tile(range1, (self.dims[2] *
                                                        self.dims[3]))\
                                         .reshape(self.dims[3], self.dims[2],
                                                  self.dims[1])
                    cart[:, :, :, 1] = np.tile(range2, (self.dims[1] *
                                                        self.dims[3]))\
                                         .reshape(self.dims[3], self.dims[1],
                                                  self.dims[2])\
                                         .transpose(0, 2, 1)
                    cart[:, :, :, 2] = np.tile(range3, (self.dims[1] *
                                                        self.dims[2]))\
                                         .reshape(self.dims[2], self.dims[1],
                                                  self.dims[3])\
                                        .transpose(2, 0, 1)
                    radius = np.sqrt(np.square(cart[:, :, :, 0]) +
                                     np.square(cart[:, :, :, 1]) +
                                     np.square(cart[:, :, :, 2]))
                    out = stopVal * np.ones([self.dims[3], self.dims[2],
                                            self.dims[1]])
            if function == 0:  # elipse
                windIdx = radius <= 1.0
                passIdx = radius <= (1.0 - window)
                out[windIdx] = (stopVal + ((1.0 - radius[windIdx]) / window) *
                                (passVal - stopVal))
                out[passIdx] = passVal

            elif function == 1:  # rectangle
                windIdx = np.amax(abs(cart), -1) <= 1.0
                passIdx = np.amax(abs(cart), -1) <= (1.0 - window)
                out[windIdx] = (stopVal +
                                ((1.0 - np.amax(abs(cart[windIdx]), -1)) /
                                window) * (passVal - stopVal))
                out[passIdx] = passVal

            elif function == 2:  # hanning
                windIdx = radius <= 1.0
                passIdx = radius <= (1.0 - window)
                func = 0.5 * (1.0 - np.cos(np.pi * (1.0 - radius[windIdx]) /
                                           window))
                out[windIdx] = stopVal + func * (passVal - stopVal)
                out[passIdx] = passVal

            elif function == 3:  # gaussian
                func = np.exp(-np.square(radius))
                out[...] = stopVal + func * (passVal - stopVal)

            elif function == 4:  # sinc
                windIdx = radius != 0.0
                func = np.pi * radius[windIdx]
                out[windIdx] = passVal * np.sin(func) / func
                out[radius == 0.0] = passVal

                # JGP Now window sinc function
                windIdx = radnorm*radius <= 1.0
                passIdx = radnorm*radius <= (1.0 - window)
                out2 = 0.*out
                func = 0.5 * (1.0 - np.cos(np.pi * (1.0 - radnorm*radius[windIdx]) / window))
                out2[windIdx] = func
                out2[passIdx] = 1.
                out *= out2

            elif function == 5:  # poly
                if ndim == 1:
                    i = cart[:,0] * self.widths[1]
                    j = 0
                    k = 0
                elif ndim == 2:
                    i = cart[:, :, 0] * self.widths[1]
                    j = cart[:, :, 1] * self.widths[2]
                    k = 0
                else:
                    i = cart[:, :, :, 0] * self.widths[1]
                    j = cart[:, :, :, 1] * self.widths[2]
                    k = cart[:, :, :, 2] * self.widths[3]
                ifunc = 0.5 * polyi * i + 0.25 * polyi2 * np.square(i)
                jfunc = 0.5 * polyj * j + 0.25 * polyj2 * np.square(j)
                kfunc = 0.5 * polyk * k + 0.25 * polyk2 * np.square(k)
                out[...] = polyC + ifunc + jfunc + kfunc

            elif function == 6:  # riesz
                if not same_dim:
                    windIdx = np.amax(abs(cart), -1) <= 1.0
                    passIdx = np.amax(abs(cart), -1) <= (1.0 - window)
                    func = (np.amax(abs(cart[windIdx]), -1) - (1.0 - window))
                else:
                    windIdx = radius <= 1.0
                    passIdx = radius <= (1.0 - window)
                    func = radius[windIdx] - (1.0 - window)
                if window != 0.0:
                    func = np.square(func / window) * (passVal - stopVal)
                else:
                    func = 0.0
                out[windIdx] = passVal - func
                out[passIdx] = passVal

            else:  # noise
                if(fixedSeed):
                    np.random.seed(self.getVal('Seed'))
                else:
                    np.random.seed()
                if ndim == 1:
                    out = std * np.random.randn(self.dims[1])
                    if complex:
                        out = out + 1j * std * np.random.randn(self.dims[1])
                elif ndim == 2:
                    out = std * np.random.randn(self.dims[2], self.dims[1])
                    if complex:
                        out = out + 1j * std * np.random.randn(self.dims[2], 
                                                               self.dims[1])
                else:
                    out = std * np.random.randn(self.dims[3], self.dims[2],
                                                self.dims[1])
                    if complex:
                        out = out + 1j * std * np.random.randn(self.dims[3], 
                                                    self.dims[2],self.dims[1])
        else:
            out = None

        self.setData('out', out)
        return 0

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Nick Zwart
# Date: 2012oct28

import gpi
import sys
from textwrap import dedent
import traceback


class ExternalNode(gpi.NodeAPI):
    """This node provides a simple code input interface and editor for
    generating Python code to be executed in the node compute() on-the-fly.
    The editor includes syntax highlighting.  The four InPorts provided are
    labeled: 'in1', 'in2', 'in3', and 'in4' (similarly for the OutPorts).  This
    code is run in a Python exec statement and therefore has associated
    limitations (e.g. return statements).  New widgets cannot be added to the
    widget interface, however, packages that contain UI elements can be spawned
    using the Execution-Type 'Main Loop'.
    """

    def execType(self):
        # skip execType recursion by getting val from widget directly
        op = self.getWidget('Execution Type').get_val()
        if op == 0:
            return gpi.GPI_THREAD
        if op == 1:
            return gpi.GPI_PROCESS
        if op == 2:
            return gpi.GPI_APPLOOP

    def initUI(self):
        self.procType = gpi.GPI_APPLOOP

        self.addWidget('TextBox', 'Status', val='Ready.')

        code = '\n'.join((
            "in1 = self.getData('in1')",
            "in2 = self.getData('in2')",
            "in3 = self.getData('in3')",
            "in4 = self.getData('in4')",
            "",
            "import numpy as np",
            "# your code here...",
            "",
            "self.setData('out1', None)",
            "self.setData('out2', None)",
            "self.setData('out3', None)",
            "self.setData('out4', None)"))

        # Widgets
        self.addWidget('TextEdit', 'Python Code', val=code)
        self.addWidget('PushButton', 'Compute', button_title='Compute')
        self.addWidget('ExclusiveRadioButtons', 'Execution Type',
                       buttons=['Thread', 'Process', 'Main Loop'],
                       val=1, collapsed=True)

        # IO Ports
        self.addInPort('in1', 'PASS', obligation=gpi.OPTIONAL)
        self.addInPort('in2', 'PASS', obligation=gpi.OPTIONAL)
        self.addInPort('in3', 'PASS', obligation=gpi.OPTIONAL)
        self.addInPort('in4', 'PASS', obligation=gpi.OPTIONAL)
        self.addOutPort('out1', 'PASS')
        self.addOutPort('out2', 'PASS')
        self.addOutPort('out3', 'PASS')
        self.addOutPort('out4', 'PASS')

    def compute(self):

        import sys

        code = str(self.getVal('Python Code'))

        try:
            self.setAttr('Status', val='Running input user code.')
            exec(code)
            self.setAttr('Status', val='User code executed successfully.')
            self.setAttr('Status', val='Ready.')

        except:
            self.log.warn("ERROR: User code failed to execute!")
            self.setData('out1', None)
            self.setData('out2', None)
            self.setData('out3', None)
            self.setData('out4', None)

            txt = "ERROR: User code failed to execute:\n\n"
            txt += traceback.format_exc()

            self.setAttr('Status', val=txt)

        return(0)



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Dallas Turley

import gpi
import numpy as np

class ExternalNode(gpi.NodeAPI):
    """Uses the numpy PIL interface for writing arrays. File types are specified in the bottom of the file browser
    Currently lossless TIFF and compressed JPEG are fully supported.

    INPUT - Image array to write. This node is designed to write to file the output of the ImageDisplay node.
        Image can be MxNx3 or MxNx4
        If the shape is MxNx3, the array will store the RGB bands as the last dimension
        If the shape is MxNx4, the array will store RGBA as the last dimension

    WIDGETS:
    File Browser - button to launch file browser, and typein widget, to give pathname for output file
      ***NOTE: If no extension (filetype) is specified, the input will be written as a .tiff file.
      Automatically appending the filetype based on the chosen file filter (in the Save File dialog box)
      has yet to be implemented. 
    Write Mode - write at any event, or write only with new filename
    Write Now - write right now
    """
    
    def execType(self):
        return gpi.GPI_THREAD

    def initUI(self):
       # Widgets
        self.addWidget(
            'SaveFileBrowser', 'File Browser', button_title='Browse',
            caption='Save Image', filter='tiff (*.tiff);;jpg (*.jpg);;all (*)')
        self.addWidget('PushButton', 'Write Mode', button_title='Write on New Filename', toggle=True)
        self.addWidget('PushButton', 'Write Now', button_title='Write Right Now', toggle=False)

        # IO Ports
        self.addInPort('in', 'NPYarray', dtype=np.uint8, ndim=3)

        # store for later use
        self.URI = gpi.TranslateFileURI

    def validate(self):
        fname = self.URI(self.getVal('File Browser'))
        self.setDetailLabel(fname)

        if self.getVal('Write Mode'):
            self.setAttr('Write Mode', button_title="Write on Every Event")
        else:
            self.setAttr('Write Mode', button_title="Write on New Filename")

        data = self.getData('in')
        if data.ndim != 3:
            self.log.warn("data must be 3D (RGB or RGBA in last dim)")
            return 1
        if data.shape[-1] not in (3,4):
            self.log.warn("The last dimension for 3D data must be 3 or 4")
            return 1

        return 0

    def compute(self):

        import numpy as np
        from PIL import Image

        data = self.getData('in').copy()

        if data.ndim == 3:
            data[...,(0,2)] = data[...,(2,0)]

        #Note: scipy.misc.imsave() uses the Python Imaging Library (PIL) which automatically 
        #chooses the format in which to save files based on the file extension (.jpg, .tif, etc)

        if self.getVal('Write Mode') or self.getVal('Write Now') or ('File Browser' in self.widgetEvents()):

            fname = self.URI(self.getVal('File Browser'))
            if fname == '':
                return 0

            #the following are valid, supported image filetype extensions
            ext_list = ('.jpg', '.tiff')
            if (fname.lower().endswith(ext_list)):
                pass
            else:
                fname += '.tiff'

            img = Image.fromarray(data)
            img.save(fname)
            self.log.info("File Written : " +str(fname))

        return(0)



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Nick Zwart
# Date: 2012dec02
# Brief: A module for reading Philips Scanner data sets.

import gpi
import numpy as np
from gpi import QtWidgets


# WIDGET
class FFTW_GROUP(gpi.GenericWidgetGroup):
    """A combination of SpinBoxes, DoubleSpinBoxes, and PushButtons
    to form a unique widget suitable for FFT options on dimensions.
    """
    valueChanged = gpi.Signal()

    def __init__(self, title, parent=None):
        super().__init__(title, parent)

        self._val = {}
        self._val['compute'] = False
        self._val['length'] = 1
        self._val['in_len'] = 1  # the original array length

        self.pb = gpi.BasicPushButton()
        self.pb.set_toggle(True)

        self.db = gpi.BasicDoubleSpinBox()  # factor
        self.db.set_label('factor:')
        self.db.set_min(0.001)
        self.db.set_max(gpi.GPI_FLOAT_MAX)
        self.db.set_decimals(3)
        self.db.set_singlestep(0.001)
        self.db.set_val(1)

        self.sb = gpi.BasicSpinBox()  # length
        self.sb.set_label('length:')
        self.sb.set_min(1)
        self.sb.set_val(1)
        self.sb.set_max(gpi.GPI_INT_MAX)

        self.db.valueChanged.connect(self.factChange)
        self.sb.valueChanged.connect(self.lenChange)
        self.pb.valueChanged.connect(self.compChange)

        vbox = QtWidgets.QHBoxLayout()
        vbox.addWidget(self.pb)
        vbox.addWidget(self.db)
        vbox.addWidget(self.sb)
        vbox.setStretch(0, 0)
        vbox.setStretch(1, 0)
        vbox.setStretch(2, 0)
        vbox.setContentsMargins(0, 0, 0, 0)  # we don't need margins here
        vbox.setSpacing(0)
        self.setLayout(vbox)

    # setters
    def set_val(self, val):
        """A python-dict containing: in_len, length, and compute parms. """
        sig = False
        if 'in_len' in val:
            # otherwise this would change every time compute() was called
            if self._val['in_len'] != val['in_len']:
                self._val['in_len'] = val['in_len']
                fact = self.db.get_val()  # set len based on factor
                fact *= self._val['in_len']
                self.setLenQuietly(int(fact))
                self._val['length'] = int(fact)
        if 'length' in val:
            self._val['length'] = val['length']
            self.setLenQuietly(val['length'])
            self.setFactQuietly(float(val[
                                'length'])/float(self._val['in_len']))
            sig = True
        if 'compute' in val:
            self._val['compute'] = val['compute']
            self.setCompQuietly(val['compute'])
            sig = True
        if sig:
            self.valueChanged.emit()

    def set_reset(self):
        """An override that communicates with the embedded pushbutton. """
        self.pb.set_reset()

    # getters
    def get_val(self):
        return self._val

    # support
    def factChange(self, val):
        self._val['length'] = int(self._val['in_len']*val)
        self.setLenQuietly(self._val['length'])
        self.valueChanged.emit()

    def lenChange(self, val):
        self._val['length'] = val
        self.setFactQuietly(float(val)/float(self._val['in_len']))
        self.valueChanged.emit()

    def compChange(self, val):
        self._val['compute'] = val
        self.valueChanged.emit()
        if val:
            self.pb.set_button_title('ON')
        else:
            self.pb.set_button_title('')

    def setFactQuietly(self, val):
        self.db.blockSignals(True)
        self.db.set_val(val)
        self.db.blockSignals(False)

    def setLenQuietly(self, val):
        self.sb.blockSignals(True)
        self.sb.set_val(val)
        self.sb.blockSignals(False)

    def setCompQuietly(self, val):
        self.pb.blockSignals(True)
        self.pb.set_val(val)
        self.pb.blockSignals(False)


class ExternalNode(gpi.NodeAPI):
    """A module that implements the FFTW C++ package.
    Cropping and zero-padding only work on transformed dimensions.

    INPUT - data to be transformed, can be real or complex.  DC is assumed to be at index N/2 (starting at 0)

    OUTPUT - transformed data, complex.  DC is at index N/2 (starting at 0)

    WIDGETS:
    Dimension i - button turns off/on tranform in ith dimension
                  factor and length are redundant parameters, length is the output dimension size
                  factor = length/(input dimension size)
                  factors < 1 result in cropping data before transformation
                  factors > 1 result in zero-padding before transformation
    compute - compute
    direction - select whether you want a Forward or Inverse FFT
    """

    def execType(self):
        # default executable type
        # return GPI_THREAD
        return gpi.GPI_PROCESS
        # return GPI_APPLOOP

    def initUI(self):

        # Widgets
        self.dim_base_name = 'Dimension['
        self.ndim = 10  # underlying c-code is only 10-dim
        for i in range(self.ndim):
            self.addWidget('FFTW_GROUP', self.dim_base_name+str(-i-1)+']')

        self.addWidget('PushButton', 'inverse', toggle=True)
        self.addWidget('PushButton', 'compute', toggle=True)

        # IO Ports
        self.addInPort('in', 'NPYarray', obligation=gpi.REQUIRED)
        self.addOutPort('out', 'NPYarray', dtype=np.complex64)

    def validate(self):
        '''update the widget bounds based on the input data
        '''

        # only update bounds if the 'in' port changed.
        if 'in' in self.portEvents():

            data = self.getData('in')

            # visibility and bounds
            for i in range(self.ndim):
                if i < len(data.shape):
                    val = {'in_len': data.shape[-i-1]}
                    self.setAttr(self.dim_base_name+str(-i-1)+']',
                                      visible=True, val=val)
                else:
                    self.setAttr(self.dim_base_name+str(-i-1)+']',
                                      visible=False)

        # only change bounds if the 'direction' widget changed.
        if 'direction' in self.widgetEvents():
            direction = self.getVal('direction')
            if direction:
                self.setAttr('direction', button_title="INVERSE")
            else:
                self.setAttr('direction', button_title="FORWARD")

        return(0)

    def compute(self):

        data = self.getData('in')
        data = np.require(data, dtype=np.complex64, requirements='C')

        if self.getVal('compute'):

            # build the fft.fft argument list
            kwargs = {}

            # Direction | 0:FWD, 1:BKWD
            if self.getVal('inverse'):
                kwargs['dir'] = 1
            else:
                kwargs['dir'] = 0

            out_dims = np.array([], np.int64)

            # load up the dimension args
            for i in range(self.ndim):
                kwargs['dim'+str(i+1)] = 0
                val = self.getVal(self.dim_base_name+str(-i-1)+']')
                if i < len(data.shape):
                    out_dims = np.append(out_dims, val['length'])
                    if val['compute']:
                        kwargs['dim'+str(i+1)] = 1


            # import in thread to save namespace
            import gpi_core.math.fft as ft

            out = ft.fftw(data, out_dims, **kwargs)

            self.setData('out', out)

        return(0)



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pype
# Date: 2013 Aug

import gpi
import numpy as np

class ExternalNode(gpi.NodeAPI):
    """Inverse Gridding module for Post-Cartesian Data using DFT - works with 2D data
    This is to create exact k-space data corresponding to any image, often for simulation and testing

    INPUTS:
    image - k-space complex data - if not supplied, Grid uses "1" for all of its data
    coords - k-space coordinates, normalized in units of "1/resolution", i.e. ranging from -0.5 to 0.5
               Last dimension must be 2 or 3 (corresponding to kx/ky or kx/ky/kz, respectively)
               Coordinates at the very edge of gridded k-space then have values of -/+ 0.5

    OUTPUTS:
    out - gridded data, same dimensions as coords
    
    WIDGETS:
    Eff Mtx - effective matrix of coords (specifies Nyquist distance)
    """

    def execType(self):
        return gpi.GPI_PROCESS

    def initUI(self):

        # Widgets
        self.addWidget('SpinBox','Eff Mtx',min=10,val = 100)

        # IO Ports
        self.addInPort('data', 'NPYarray', ndim=2, dtype=[np.complex64, np.complex128])
        self.addInPort('coords', 'NPYarray', dtype=[np.float32, np.float64])
        self.addInPort('weighting', 'NPYarray', dtype=[np.float32, np.float64], obligation=gpi.OPTIONAL)
        self.addOutPort('image', 'NPYarray', dtype= np.complex128)

    def validate(self):
        crds = self.getData('coords')

        if crds.shape[-1] != 2:
          self.log.warn(" Error: only 2 dimensional gridding implemented")
          return 1
        
        return 0

    def compute(self):

        import numpy as np
        import gpi_core.gridding.dft as dft

        effmtx = self.getVal('Eff Mtx')
        data = self.getData('data').astype(np.complex128)
        crds = self.getData('coords').astype(np.float64)
        wghts = self.getData('weighting')
        newcrds = np.reshape(crds,(crds[...,0].size,2))

        if wghts is not None:
          if wghts.ndim != crds.ndim - 1:
            self.log.warn("# of dimensions of weights must match crds")
            return 1
          else: 
            for i in range(crds.ndim-1):
              if crds.shape[i] != wghts.shape[i]:
                self.log.warn("sizes of weights and coords don't match")
                return 1

        # Make wghts or reshape it
        if wghts is None:
            wghts = np.ones((crds[...,0].size,), dtype=np.float32)
        else:
            # Reshape wghts to be 2 dimensional
            wghts = np.reshape(wghts.astype(np.float32),(crds[...,0].size,))

        #outdim = np.array([crds[...,0].size],dtype=np.int64)
        outdim = np.array([effmtx,effmtx],dtype=np.int64)

        out = dft.dft_grid(data,newcrds,outdim,effmtx,wghts.astype(np.float64))

        #out = np.reshape(out,crds[...,0].shape)
        self.setData('image', out)

        return(0)



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Ryan Robison
# Date: 2015aug25

import gpi
from gpi import QtWidgets


class GPITabBar(QtWidgets.QTabBar):
    currentChanged = gpi.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self._labels_at_press = None

    def clearTabs(self):
        while self.count() > 0:
            self.removeTab(0)

    def addTabsFromList(self, names):
        for i in range(len(names)):
            self.addTab(str(names[i]))

    def labels(self):
        return [str(self.tabText(i)) for i in range(self.count())]

    def mousePressEvent(self, event):
        self._labels_at_press = self.labels()
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)
        if self.labels() != self._labels_at_press:
            self.currentChanged.emit()

class OrderButtons(gpi.GenericWidgetGroup):
    """A set of reorderable tabs."""
    valueChanged = gpi.Signal()

    def __init__(self, title, parent=None):
        super().__init__(title, parent)

        # at least one button
        wdgLayout = QtWidgets.QGridLayout()
        self.wdg = GPITabBar()
        self.wdg.addTab('0')
        self.wdg.setMovable(True)
        self.wdg.currentChanged.connect(self.valueChanged.emit)
        # layout
        wdgLayout.addWidget(self.wdg)
        self.setLayout(wdgLayout)

    # setters
    def set_val(self, names):
        """list(str,str,...) | A list of labels (e.g. ['b1', 'b2',...])."""
        self.wdg.clearTabs()
        self.wdg.addTabsFromList(names)

    # getters
    def get_val(self):
        return self.wdg.labels()


class ExternalNode(gpi.NodeAPI):
    """Peform transpose operations on an array.
    INPUT - input array
    OUTPUT - output array

    WIDGETS:
    Info: - information on size, etc of input and output array
    Transpose - transpose any dimensions
    """

    def initUI(self):

        # Widgets
        self.addWidget('TextBox', 'Info:')
        self.addWidget('PushButton', 'Transpose', toggle=True, val=1)
        self.addWidget('OrderButtons', 'Dimension Order')

        # IO Ports
        self.addInPort('in', 'NPYarray', obligation=gpi.REQUIRED)
        self.addOutPort('out', 'NPYarray')

    def validate(self):
        data = self.getData('in')
        order = self.getVal('Dimension Order')

        # the widget doesn't have the same dims as the data then reset the
        # widget with the new length
        if len(order) != data.ndim:
            self.setAttr('Dimension Order', quietval=[str(i) for i in range(data.ndim)])

        # automatically transpose if ndim = 2
        if data.ndim < 3:
            self.setAttr('Dimension Order', visible=False)
        else:
            self.setAttr('Dimension Order', visible=True)

        return 0

    def compute(self):

        data = self.getData('in')
        transpose = self.getVal('Transpose')
        order = self.getVal('Dimension Order')

        # display info
        basic_info = "Input Dimensions: "+str(data.shape)+"\n"

        # setup the transpose indices (automatically transpose if ndim = 2)
        trans_ind = data.ndim*[0]
        if data.ndim > 2:
            for i in range(len(order)):
                trans_ind[i] = int(order[i])
        else:
            trans_ind = [1, 0]

        # compute
        if transpose and data.ndim > 1:
            out = data.transpose(trans_ind)
        else:
            out = data

        # updata info
        info = basic_info+"Output Dimensions: "+str(out.shape)+"\n"
        self.setAttr('Info:', val = info)
        self.setData('out', out)

        return 0



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Ryan Robison
# Date: 2015nov13

import numpy as np
import gpi

class ExternalNode(gpi.NodeAPI):
    """Shifts data along each dimension by user specified amount.
    Shift can be linear or circular.
    INPUT - input array
    OUTPUT - output array

    WIDGETS:
    """

    def initUI(self):

        # Widgets
        self.dim_base_name = 'Dimension['
        self.ndim = 13  # 13 for now
        self.addWidget('ExclusivePushButtons', 'Type', 
                buttons=['Linear', 'Circular'], val=0)
        for i in range(self.ndim):
            self.addWidget('SpinBox', self.dim_base_name+str(-i-1)+']', val=0)

        # IO Ports
        self.addInPort('in', 'NPYarray', obligation=gpi.REQUIRED)
        self.addOutPort('out', 'NPYarray')

    def validate(self):
        '''update the widgets based on the input arrays
        '''

        data = self.getData('in')

        # visibility and bounds
        for i in range(self.ndim):
            if i < len(data.shape):
                shift = self.getVal(self.dim_base_name+str(-i-1)+']')
                self.setAttr(self.dim_base_name+str(-i-1)+']',
                        visible=True, val=shift)
            else:
                self.setAttr(self.dim_base_name+str(-i-1)+']',
                        visible=False)

        return(0)

    def compute(self):

        data = self.getData('in')
        out = data
        stype = self.getVal('Type')

        for i in range(data.ndim):
            shift = self.getVal(self.dim_base_name+str(-i-1)+']')

            if stype == 0:
                if shift <= 0:
                    shift = max(0, -shift)
                    del_range = list(range(shift))
                    insert_vals = [data.shape[-i-1]] * shift
                    out = np.insert(out, insert_vals, 0, axis = -i-1)
                    out = np.delete(out, del_range, axis = -i-1)
                else:
                    shift = min(shift, data.shape[-i-1])
                    insert_vals = [0] * shift
                    del_range = list(range(data.shape[-i-1] - shift,
                                      data.shape[-i-1] + 1))
                    out = np.delete(out, del_range, axis = -i-1)
                    out = np.insert(out, insert_vals, 0, axis = -i-1)

            if stype == 1:
                out = np.roll(out, shift, axis = -i-1)

        self.setData('out', out)

        return(0)

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Nick Zwart
# Date 2013 sep 01

# NOTE THIS IS A SPECIAL MODULE, AND WILL LIKELY BE SUBSTANTIALLY REPLACED AT LATER DATE
# --> YOU SHOULD NOT USE THIS MODULE AS A TEMPLATE

import sys
import traceback
import numpy as np

import gpi
import gpi.GLObjects as glo
import gpi.logger
from gpi.defines import getKeyboardModifiers, printMouseEvent
from gpi.numpyqt import qimage2numpy
# start logger for this module
log = gpi.logger.manager.getLogger(__name__)

from gpi import QtCore, QtGui, Qimport, QtWidgets, QtOpenGL, QT_API_NAME

# TODO: QtOpenGL is deprecated in recent Qt 5
# could use newer QtWidgets.QOpenGLWidget, etc. instead

try:
    from OpenGL import GL, GLU, GLUT
    import OpenGL.GL
    import OpenGL.GLU
    import OpenGL.GLUT

except ImportError:
    app = QtWidgets.QApplication(sys.argv)
    QtWidgets.QMessageBox.critical(None, "OpenGL grabber",
                               "PyOpenGL must be installed to run this example.")
    raise


class GPIGLWidget(QtOpenGL.QGLWidget):
    xRotationChanged = gpi.Signal(int)
    yRotationChanged = gpi.Signal(int)
    zRotationChanged = gpi.Signal(int)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.xRot = 0
        self.yRot = 0
        self.zRot = 0
        self._vScale = 1.0
        self._xPan = 0.0
        self._yPan = 0.0

        # a list of object descriptions held in dictionaries
        self._GPI_glList = glo.ObjectList()

        # trac number of exceptions thrown in problem code
        self._glError_cnt = 0

        # cache objects in a gl list
        self._glList_cache = None
        self._glList_nonCacheable = []

        # lighting
        self._default_lightPos = [0.0, 0.0, 10.0, 1.0]
        self._lightPos = self._default_lightPos[:]

        # built-in options
        self._blend = 0
        self._test = 0
        self._polyfill = 0
        self._accum_antialiasing = 0

        # self.setMouseTracking(True)

        self.checkFormat()

    def checkFormat(self):
        msg = "\n\taccum buffer:" + str(self.format().accum()) + "\n"
        msg += "\talpha buffer:" + str(self.format().alpha()) + "\n"
        msg += "\tdepth buffer:" + str(self.format().depth()) + "\n"
        msg += "\tdirectRendering:" + \
            str(self.format().directRendering()) + "\n"
        msg += "\tdoubleBuffer:" + str(self.format().doubleBuffer()) + "\n"
        msg += "\thasOverlay:" + str(self.format().hasOverlay()) + "\n"
        msg += "\tplane:" + str(self.format().plane()) + "\n"
        msg += "\trgba:" + str(self.format().rgba()) + "\n"
        log.node(msg)

    def setGPIglList(self, lst):
        if isinstance(lst, glo.ObjectList):
            self._GPI_glList = lst

    def getGPIglList(self):
        return self._GPI_glList

    def __del__(self):
        # cannot guarantee that the underlying object hasn't been deleted
        # before this context is made current
        try:
            self.makeCurrent()
        except:
            log.node(traceback.format_exc())
        self.resetGLCache()

    def setViewScale(self, s):
        self._vScale += s
        self.updateGL()

    def setXRotation(self, angle):
        self.normalizeAngle(angle)

        if angle != self.xRot:
            self.xRot = angle
            self.xRotationChanged.emit(angle)
            self.updateGL()

    def setYRotation(self, angle):
        self.normalizeAngle(angle)

        if angle != self.yRot:
            self.yRot = angle
            self.yRotationChanged.emit(angle)
            self.updateGL()

    def setZRotation(self, angle):
        self.normalizeAngle(angle)

        if angle != self.zRot:
            self.zRot = angle
            self.zRotationChanged.emit(angle)
            self.updateGL()

    def initializeGL(self):

        # from basic gear pos
        GL.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, self._lightPos)
        GL.glEnable(GL.GL_LIGHTING)
        GL.glEnable(GL.GL_LIGHT0)
        GL.glEnable(GL.GL_DEPTH_TEST)
        GL.glEnable(GL.GL_NORMALIZE)

        # set scene to black
        GL.glClearColor(0.0, 0.0, 0.0, 0.0)

        # cache input objects
        self._clipON = False
        # self.instantiateRefs()
        self.cacheGLCommands()

    def paintGL(self):

        GL.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, self._lightPos)

        if self._blend:

            GL.glCullFace(GL.GL_BACK)
            GL.glEnable(GL.GL_CULL_FACE)
            GL.glBlendFunc(GL.GL_SRC_ALPHA_SATURATE, GL.GL_ONE)

            # something is not properly initialized for this to run
            # correctly at the beginning.
            try:
                GL.glClear(GL.GL_COLOR_BUFFER_BIT)
            except:
                log.node(traceback.format_exc())
                self._glError_cnt += 1

            GL.glEnable(GL.GL_BLEND)
            GL.glEnable(GL.GL_POLYGON_SMOOTH)
            GL.glDisable(GL.GL_DEPTH_TEST)
        else:
            GL.glDisable(GL.GL_CULL_FACE)

            # something is not properly initialized for this to run
            # correctly at the beginning.
            try:
                GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)
            except:
                log.node(traceback.format_exc())
                self._glError_cnt += 1

            GL.glDisable(GL.GL_BLEND)
            GL.glDisable(GL.GL_POLYGON_SMOOTH)
            GL.glEnable(GL.GL_DEPTH_TEST)

        if self._polyfill == 2:
            GL.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_POINT)
        elif self._polyfill == 1:
            GL.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_LINE)
        else:
            GL.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_FILL)

        if self._accum_antialiasing:
            GL.glClear(GL.GL_ACCUM_BUFFER_BIT)
            cnt = 3
            mult = cnt * 2
            mult *= mult
            invmult = 1.0 / mult
            for i in range(-cnt, cnt):
                for j in range(-cnt, cnt):
                    # jitter and paint
                    GL.glPushMatrix()
                    GL.glTranslatef(i * 0.00511, j * 0.00511, 0.0)
                    self.paintScene()
                    GL.glPopMatrix()

                    # accum
                    GL.glAccum(GL.GL_ACCUM, invmult)
            GL.glAccum(GL.GL_RETURN, 1.0)
        else:
            self.paintScene()

        if self._test:
            pass
        else:
            pass

        # force all commands to complete
        GL.glFlush()

    def paintScene(self):
        '''Does the actuall object rendering calls.
        '''
        # rotate view
        GL.glPushMatrix()
        GL.glRotatef(self.xRot / 16.0, 1.0, 0.0, 0.0)
        GL.glRotatef(self.yRot / 16.0, 0.0, 1.0, 0.0)
        GL.glRotatef(self.zRot / 16.0, 0.0, 0.0, 1.0)

        # pan
        mtx = np.array(GL.glGetDoublev(GL.GL_MODELVIEW_MATRIX))
        A = mtx[0:3, 0:2]
        x = np.transpose(np.array([[self._xPan, self._yPan]]))
        b = np.dot(A, x)
        GL.glTranslatef(b[0], b[1], b[2])

        # zoom
        GL.glScale(self._vScale, self._vScale, self._vScale)

        # render all non-cacheable first
        if len(self._glList_nonCacheable):
            for desc in self._glList_nonCacheable:
                desc.run()

        # use cached list for mouse interactions
        try:
            GL.glCallList(self._glList_cache)
        except:
            log.node(traceback.format_exc())
            self._glError_cnt += 1

        # rotate view
        GL.glPopMatrix()

    def resizeGL(self, width, height):
        side = min(width, height)
        if side < 0:
            return

        GL.glViewport((width - side) // 2, (height - side) // 2, side, side)

        GL.glMatrixMode(GL.GL_PROJECTION)
        GL.glLoadIdentity()
        GL.glFrustum(-1.0, +1.0, -1.0, 1.0, 5.0, 60.0)
        GL.glMatrixMode(GL.GL_MODELVIEW)
        GL.glLoadIdentity()
        GL.glTranslatef(0.0, 0.0, -40.0)

    def resetGLCache(self):
        '''reset both cacheable and non-cacheable items
        '''
        if self._glList_cache:
            GL.glDeleteLists(self._glList_cache, 1)
        self._glList_cache = None
        self._glList_nonCacheable = []

    def instantiateRefs(self):
        '''Run thru all objects calling their instantiateLibs() method.
        '''
        for desc in self._GPI_glList:
            desc.instantiateRefs()
            if type(desc) is glo.ClipPlane:
                print("enable clipping", desc.getPlaneNumTr())
                GL.glEnable(desc.getPlaneNumTr())

    def cacheGLCommands(self):
        '''Cache object commands.
        '''
        if self._glList_cache:
            self.resetGLCache()

        self._glList_cache = GL.glGenLists(1)
        GL.glNewList(self._glList_cache, GL.GL_COMPILE)

        # cache all commands in obj list
        for desc in self._GPI_glList.getCacheableList():
            desc.run()

        # special objects
        for plane, desc in self._GPI_glList.getClipPlanes().items():
            print('render: ' + plane)
            desc.run()

        GL.glEndList()

        # get all non-cacheables
        self._glList_nonCacheable = self._GPI_glList.getNonCacheableList()
        if len(self._glList_nonCacheable):
            for desc in self._glList_nonCacheable:
                desc.setGLWidgetRef(self)

    def wheelEvent(self, event):
        try:
            # PyQt4
            delta = event.delta()
        except AttributeError:
            # PyQt5
            delta = event.angleDelta().y()
        if delta > 0:
            self.setViewScale(0.1)
        else:
            self.setViewScale(-0.1)

    def mousePressEvent(self, event):
        self.lastPos = event.pos()

    def mouseMoveEvent(self, event):
        dx = event.x() - self.lastPos.x()
        dy = event.y() - self.lastPos.y()

        printMouseEvent(event)
        modifiers = getKeyboardModifiers()
        modmidbutton_event = (event.buttons() & QtCore.Qt.RightButton
                              and modifiers & QtCore.Qt.ShiftModifier)

        if event.buttons() & QtCore.Qt.LeftButton:
            self._xPan += dx * 0.05
            self._yPan += -dy * 0.05
            self.updateGL()
        elif modmidbutton_event:
            self.setXRotation(self.xRot + 8 * dy)
            self.setYRotation(self.yRot + 8 * dx)
        elif event.buttons() & QtCore.Qt.MidButton:
            self._lightPos[0] += dx * 0.05
            self._lightPos[1] += -dy * 0.05
            self.updateGL()
        elif event.buttons() & QtCore.Qt.RightButton:
            self.setXRotation(self.xRot + 8 * dy)
            self.setZRotation(self.zRot + 8 * dx)

        self.lastPos = event.pos()

    def resetViewingWindow(self):
        '''reset pan, zoom, rotate
        '''
        self._xPan = 0.0
        self._yPan = 0.0
        self._vScale = 1.0
        self.xRot = 0.0
        self.yRot = 0.0
        self.zRot = 0.0
        self._lightPos = self._default_lightPos[:]

    def xRotation(self):
        return self.xRot

    def yRotation(self):
        return self.yRot

    def zRotation(self):
        return self.zRot

    def normalizeAngle(self, angle):
        while (angle < 0):
            angle += 360 * 16

        while (angle > 360 * 16):
            angle -= 360 * 16

    def setBlend(self, val):
        self._blend = val
        self.updateGL()

    def setTest(self, val):
        self._test = val
        self.updateGL()

    def setPolyFill(self, val):
        self._polyfill = val
        self.updateGL()

    def setAntiAliasing(self, val):
        self._accum_antialiasing = val
        self.updateGL()


class OpenGLWindow(gpi.GenericWidgetGroup):

    """Provides an embedded GL window.
    """
    valueChanged = gpi.Signal()

    def __init__(self, title, parent=None):
        if QtOpenGL is None:
            raise ImportError("QtOpenGL not available in the current Qt "
                              "package ({})".format(QT_API_NAME))
        super().__init__(title, parent)
        f = QtOpenGL.QGLFormat()
        f.setAccum(True)
        f.setDoubleBuffer(True)
        f.setRgba(True)
        f.setDepth(True)
        f.setAlpha(True)
        self.glWidget = GPIGLWidget(f)

        self.glWidgetArea = QtWidgets.QScrollArea()
        self.glWidgetArea.setWidget(self.glWidget)
        self.glWidgetArea.setWidgetResizable(True)
        self.glWidgetArea.setHorizontalScrollBarPolicy(
            QtCore.Qt.ScrollBarAlwaysOff)
        self.glWidgetArea.setVerticalScrollBarPolicy(
            QtCore.Qt.ScrollBarAlwaysOff)
        self.glWidgetArea.setSizePolicy(QtWidgets.QSizePolicy.Ignored,
                                        QtWidgets.QSizePolicy.Ignored)
        self.glWidgetArea.setMinimumSize(50, 50)

        # self.pixmapLabelArea = QtWidgets.QScrollArea()
        # self.pixmapLabelArea.setWidget(self.pixmapLabel)
        # self.pixmapLabelArea.setSizePolicy(QtWidgets.QSizePolicy.Ignored,
        #        QtWidgets.QSizePolicy.Ignored)
        # self.pixmapLabelArea.setMinimumSize(50, 50)

        xSlider = self.createSlider(self.glWidget.xRotationChanged,
                                    self.glWidget.setXRotation)
        ySlider = self.createSlider(self.glWidget.yRotationChanged,
                                    self.glWidget.setYRotation)
        zSlider = self.createSlider(self.glWidget.zRotationChanged,
                                    self.glWidget.setZRotation)

        glBlend = self.createCheckOption('PolySmooth', self.glWidget.setBlend)
        polyFill = self.createCheckOption(
            'Poly-Fill/Line/Point', self.glWidget.setPolyFill, tristate=True)

        # antialiasing requires the accumulation buffer
        enableaccum = True
        if not self.glWidget.format().accum():
            enableaccum = False
        antialiasing = self.createCheckOption(
            'AntiAliasing', self.glWidget.setAntiAliasing, initstate=0, enabled=enableaccum)

        hardwareRender = QtWidgets.QLabel()
        hardwareRender.setFrameStyle(2)
        if self.glWidget.format().directRendering():
            hardwareRender.setText('Rendering: Hardware')
        else:
            hardwareRender.setText('Rendering: Software')

        #testOption = self.createCheckOption(
        #    'Test Option', self.glWidget.setTest, tristate=False)

        # self.createActions()
        # self.createMenus()
        centralLayout = QtWidgets.QGridLayout()
        centralLayout.addWidget(self.glWidgetArea, 2, 0, 4, 4)
        # centralLayout.setColumnStretch(0,3)
        centralLayout.setRowStretch(2, 2)
        # centralLayout.addWidget(self.pixmapLabelArea, 0, 1)
        #centralLayout.addWidget(xSlider, 2, 0, 1, 2)
        #centralLayout.addWidget(ySlider, 3, 0, 1, 2)
        #centralLayout.addWidget(zSlider, 4, 0, 1, 2)
        centralLayout.addWidget(polyFill, 1, 0, 1, 1)
        centralLayout.addWidget(glBlend, 1, 1, 1, 1)
        centralLayout.addWidget(antialiasing, 1, 2, 1, 1)
        centralLayout.addWidget(hardwareRender, 0, 0, 1, 1)
        #centralLayout.addWidget(testOption, 0, 2, 1, 1)

        self.setLayout(centralLayout)

        xSlider.setValue(15 * 16)
        ySlider.setValue(345 * 16)
        zSlider.setValue(0 * 16)

        # self.setWindowTitle("Grabber")
        self.resize(400, 300)

    # setters
    def set_val(self, val):
        """set a list of GL command list interface (dict)."""
        self.glWidget.setGPIglList(val)
        self.glWidget.glInit()
        self.glWidget.updateGL()

    def set_resetView(self, val):
        """reset the viewing window"""
        self.glWidget.resetViewingWindow()

    def set_imageARGB(self, val):
        pass

    def set_imageBGRA(self, val):
        pass

    # getters
    def get_val(self):
        """get the held list of GPI-GL objects"""
        return self.glWidget.getGPIglList()

    def get_resetView(self):
        '''This is a one-shot operation, so there is nothing to get'''
        pass

    def get_imageARGB(self):
        '''Render a copy of the GL window and convert to NPY array.
        '''
        fbuff = self.glWidget.grabFrameBuffer(withAlpha=True)
        arr = qimage2numpy(fbuff)
        arr = arr[..., ::-1]
        return arr

    def get_imageBGRA(self):
        '''Render a copy of the GL window and convert to NPY array.  BGRA is
        the fastest since it native.
        '''
        fbuff = self.glWidget.grabFrameBuffer(withAlpha=True)
        arr = qimage2numpy(fbuff)
        return arr

    # support
    def renderIntoPixmap(self):
        size = self.getSize()

        if size.isValid():
            pixmap = self.glWidget.renderPixmap(size.width(), size.height())
            self.setPixmap(pixmap)

    def grabFrameBuffer(self):
        image = self.glWidget.grabFrameBuffer()
        self.setPixmap(QtGui.QPixmap.fromImage(image))

    def clearPixmap(self):
        self.setPixmap(QtGui.QPixmap())

    def about(self):
        QtWidgets.QMessageBox.about(self, "About Grabber",
                                "The <b>Grabber</b> example demonstrates two approaches for "
                                "rendering OpenGL into a Qt pixmap.")

    def createActions(self):
        self.renderIntoPixmapAct = QtWidgets.QAction("&Render into Pixmap...",
                                                 self, shortcut="Ctrl+R", triggered=self.renderIntoPixmap)

        self.grabFrameBufferAct = QtWidgets.QAction("&Grab Frame Buffer", self,
                                                shortcut="Ctrl+G", triggered=self.grabFrameBuffer)

        self.clearPixmapAct = QtWidgets.QAction("&Clear Pixmap", self,
                                            shortcut="Ctrl+L", triggered=self.clearPixmap)

        self.exitAct = QtWidgets.QAction("E&xit", self, shortcut="Ctrl+Q",
                                     triggered=self.close)

        self.aboutAct = QtWidgets.QAction("&About", self, triggered=self.about)

        self.aboutQtAct = QtWidgets.QAction("About &Qt", self,
                                        triggered=QtWidgets.qApp.aboutQt)

    def createMenus(self):
        self.fileMenu = self.menuBar().addMenu("&File")
        self.fileMenu.addAction(self.renderIntoPixmapAct)
        self.fileMenu.addAction(self.grabFrameBufferAct)
        self.fileMenu.addAction(self.clearPixmapAct)
        self.fileMenu.addSeparator()
        self.fileMenu.addAction(self.exitAct)

        self.helpMenu = self.menuBar().addMenu("&Help")
        self.helpMenu.addAction(self.aboutAct)
        self.helpMenu.addAction(self.aboutQtAct)

    def createCheckOption(self, title, setterSlot, tristate=False, initstate=0, enabled=True):
        checkbox = QtWidgets.QCheckBox(title)
        checkbox.setTristate(tristate)
        checkbox.setCheckState(initstate)
        checkbox.stateChanged.connect(setterSlot)
        if not enabled:
            checkbox.setEnabled(False)
        return checkbox

    def createSlider(self, changedSignal, setterSlot):
        slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        slider.setRange(0, 360 * 16)
        slider.setSingleStep(16)
        slider.setPageStep(15 * 16)
        slider.setTickInterval(15 * 16)
        slider.setTickPosition(QtWidgets.QSlider.TicksRight)

        slider.valueChanged.connect(setterSlot)
        changedSignal.connect(slider.setValue)

        return slider

    def setPixmap(self, pixmap):
        self.pixmapLabel.setPixmap(pixmap)
        size = pixmap.size()

        if size - QtCore.QSize(1, 0) == self.pixmapLabelArea.maximumViewportSize():
            size -= QtCore.QSize(1, 0)

        self.pixmapLabel.resize(size)

    def getSize(self):
        text, ok = QtWidgets.QInputDialog.getText(self, "Grabber",
                                              "Enter pixmap size:", QtWidgets.QLineEdit.Normal,
                                              "%d x %d" % (self.glWidget.width(), self.glWidget.height()))

        if not ok:
            return QtCore.QSize()

        regExp = QtCore.QRegExp("([0-9]+) *x *([0-9]+)")

        if regExp.exactMatch(text):
            width = regExp.cap(0).toInt()
            height = regExp.cap(1).toInt()
            if width > 0 and width < 2048 and height > 0 and height < 2048:
                return QtCore.QSize(width, height)

        return self.glWidget.size()


class ExternalNode(gpi.NodeAPI):

    """Prototype 3D GL Viewer
    MOUSE EVENTS:
      Left Button: translate object
      Middle Button: move light source
      Scroll wheel: zoom
      Right Button: rotate (hold down shift key for different axis of rotation)

    INPUT:
    GL Object List

    OUTPUT:
    2D ARGB (stored as 3D array, with last dim of length 4, uint (byte) data for 0-255 per channel)
            This is an image of what is rendered in the Viewport, useful for gluing together images to make movies

    WIDGETS:
    Viewport
      Poly-Fill Line/Point - toggles between showing polygons, lines, or points (depends on how objects are created)
      PolySmooth - for smoothing polygons, implemenation will vary
      AntiAliasing - this option needs to be debugged further
      Pixmap - shows the rendered pixmap.  In this area, the scene is manipulated as follows:
        Left Mouse Button Drag: translate object
        Middle Mouse Button Drag: translate light source
        Middle Mouse Wheel Scroll: Zoom
        Right Mouse Button Drag: Rotate Object about 2 axes.
              Hold down Shift key BEFORE selecting Right mouse button to change 2nd axis of rotation

    Reset - resets rotation, translation, zoom to original values (x axis Horizontal, y axis vertical, z axis through-plane)

    Compute/Nudge - sometimes needed to update renderer

    Set Output - makes output port continually reflect image rendered in Viewport as ARGB data

    KNOWN ISSUES:
    Still needs to be tested on many platforms, bugs likely
    On Linux platform, Z axis location occasionally seems to be improperly interpreted
    """

    def execType(self):
        return gpi.GPI_APPLOOP

    def initUI(self):

        # Widgets
        self.addWidget('OpenGLWindow', 'Viewport')
        self.addWidget('PushButton', 'Reset')
        self.addWidget('PushButton', 'Compute/Nudge')
        self.addWidget('PushButton', 'Set Output', toggle=True)

        # IO Ports
        self.addInPort(
            'GL Object Descriptions', 'GLOList', obligation=gpi.OPTIONAL)
        self.addOutPort('Rendered RGBA', 'NPYarray')

    def compute(self):

        reset = self.getVal('Reset')
        setoutput = self.getVal('Set Output')

        gpi_glist = glo.ObjectList(self.getData('GL Object Descriptions'))

        if reset:
            self.setAttr('Viewport', resetView=reset)
        self.setAttr('Viewport', val=gpi_glist) # setting the 'val' forces a redraw

        if setoutput:
            img = self.getAttr('Viewport', 'imageARGB')
            self.setData('Rendered RGBA', img)
        else:
            self.setData('Rendered RGBA', None)

        return(0)
